/*
 * Device Driver for DataTranslation DT2801
 *
 * $Id: dt2801.c,v 1.2 1996/03/13 09:34:38 martin Exp martin $
 * $Source: /home/martin/Software/DAQ/drivers/dt2801/RCS/dt2801.c,v $
 *
 *
 * $Log: dt2801.c,v $
 * Revision 1.2  1996/03/13 09:34:38  martin
 * Cleanups, strict locking for the subsystems,
 * fake mode
 *
 * Revision 1.1  1996/03/04  07:19:34  martin
 * Initial revision
 *
 *
 */

static char rcs_revision[] = "$Revision: 1.2 $";


#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/tty.h>
#include <linux/signal.h>
#include <linux/major.h>
#include <linux/malloc.h>
#include <linux/timer.h>
#include <linux/ioport.h>

#include <asm/dma.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>

#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h>
#else
#error "dt2801mod must be compiled as a module"
#endif

#include "dt2801.h"
#include "dt2801_adc.h"
#include "dt2801_dac.h"
#include "dt2801_dio.h"


Dt2801_t *dt2801 = NULL;

static int dt_timer_int = 0;
static struct timer_list dt_timer_list;

static struct wait_queue* dt2801_waitq = NULL;


static void dt2801_timer_interrupt(unsigned long unused)
{
    dprintk(( "dt2801: timer interrupt.\n" ));
    if (!dt_timer_int)  return;

    dprintk(( "dt2801: serving irq.\n" ));
    wake_up_interruptible(&dt2801_waitq);
}

static void dt2801_start_timer(void)
{
    init_timer(&dt_timer_list);
    dt_timer_list.expires = jiffies + DT2801_TIMERCYCLES;
    dt_timer_list.function = dt2801_timer_interrupt;
    dt_timer_int = 1;
    add_timer(&dt_timer_list);
}

static void dt2801_wait(void)
{
    dprintk(( "dt2801: wait.\n" ));
    dt2801_start_timer();
 
    dprintk(( "dt2801: going to sleep.\n" ));
    interruptible_sleep_on(&dt2801_waitq);
    dprintk(( "dt2801: woken up.\n" ));
    dt_timer_int = 0;
    del_timer(&dt_timer_list);

}


int dt2801_readdata(int *data);

/*
 * Stop board, and do a dummy read to clear DataOutReady-bit
 */
int dt2801_stop(void)
{
    int stat = 0;
    int ignored;

    dprintk(( "dt2801: stop.\n" ));

    outb_p(DT_C_STOP, DT_COMMAND);
    stat = dt2801_readdata(&ignored);
    dprintk(( "dt2801: ignored = %d.\n", ignored ));


    dprintk(( "dt2801: stop done.\n" ));

    return stat;
}

int dt2801_writecommand(int command);
int dt2801_readdata2(int *data);

int dt2801_clearerror(void) {

    dprintk(( "dt2801: clear error.\n" ));
    outb_p(DT_C_CLEAR_ERR, DT_COMMAND);

    return 0;
}

int dt2801_checkforerror(void)
{
    int stat = inb_p(DT_STATUS);
    int stat_tmp;
    int err;

    dprintk(( "checkforerror stat = %d .\n", stat ));
    stat_tmp = ((stat)^DT_S_DATA_IN_FULL) & DT_S_DATA_IN_FULL;
    dprintk(( "checkforerror stat = %d .\n", stat ));
    if (!stat) {
        dt2801_wait();

        stat = inb_p(DT_STATUS);
        dprintk(( "dt2801: checkerror stat = %d.\n", stat ));
        stat_tmp = ((stat)^DT_S_DATA_IN_FULL) & DT_S_DATA_IN_FULL;
        dprintk(( "dt2801: checkerror stat = %d.\n", stat ));

        if (!stat_tmp) {
            dprintk(( "dt2801: checkerror timed out. DataInFull not cleared\n" ));
        }
    }

        stat = stat & DT_S_COMPOSITE_ERROR;
        dprintk(( "checkforerror stat = %d .\n", stat ));
        if (stat)
        {

            if (dt2801->inerror)
            {
                printk("Error while doing error-handling.\n");
                return -2;
            }
            dt2801->inerror = 1;

            dt2801_stop(); 
            dt2801_writecommand(DT_C_READ_ERRREG);
            dt2801_readdata2(&err);
            printk("dt2801: Error #%d.\n", err);

            dt2801->inerror = 0;

            return -stat;
        }

        dprintk(( "checkforerror OK .\n" ));

        return 0;
}

int dt2801_readdata(int* data)
{
    static int fake_data = 0;

    int stat = 0;

    dprintk(( "dt2801: readdata.\n" ));

    if (dt2801->fake != 0) {
        fake_data = (fake_data + 1)%4096;
        *data = fake_data;
        return 0;
    }

    stat = inb_p(DT_STATUS);
    dprintk(( "dt2801: readdata stat = %d.\n", stat ));

    stat = stat & DT_S_DATA_OUT_READY;
    dprintk(( "dt2801: readdata stat = %d.\n", stat ));

    if (!stat) {
        dt2801_wait();
        dprintk(( "dt2801: readdata back.\n" ));

        stat = inb_p(DT_STATUS);
        dprintk(( "dt2801: readdata stat = %d.\n", stat ));

        stat = stat & DT_S_DATA_OUT_READY;
        dprintk(( "dt2801: readdata stat = %d.\n", stat ));

        if (!stat) {
            dprintk("dt2801: readdata timed out. DataOutReady not set\n");
            return -1;
        }

    }
    dprintk(( "dt2801: readdata ok.\n" ));


    *data = inb_p(DT_DATA_OUT);
    dprintk(( "dt2801: readdata data = %d.\n", *data ));

    return 0;
}

int dt2801_readdata2(int *data)
{
    static int fake_data = 0;
    int lb, hb;

    if (dt2801->fake != 0) {
        fake_data = (fake_data + 1)%10;
        *data = fake_data;
        return 0;
    }

    if (dt2801_readdata(&lb) != 0) {
        return -1;
    }
    if (dt2801_readdata(&hb) != 0) {
        return -1;
    }

    *data =  (hb<<8)+lb;
    return 0;
}

int dt2801_writedata(unsigned int data)
{
    int stat = 0;

    dprintk(( "dt2801: writedata data = %d.\n", data ));

    if (dt2801->fake != 0) {
        return 0;
    }

    stat = inb_p(DT_STATUS);
    dprintk(( "dt2801: writedata stat = %d.\n", stat ));

    stat = ((stat)^DT_S_DATA_IN_FULL) & DT_S_DATA_IN_FULL;
    dprintk(( "dt2801: writedata stat = %d.\n", stat ));

    if (!stat) {
        dt2801_wait();
        dprintk(( "dt2801: writedata back.\n" ));

        stat = inb_p(DT_STATUS);
        dprintk(( "dt2801: writedata stat = %d.\n", stat  ));

        stat = ((stat)^DT_S_DATA_IN_FULL) & DT_S_DATA_IN_FULL;
        dprintk(( "dt2801: writedata stat = %d.\n", stat ));

        if (!stat) {
            dprintk("dt2801: writedata timed out. DataInFull not cleared\n");
            return -1;
        }
    }
    dprintk(( "dt2801: writedata ok data = %d.\n", data ));

    outb_p(data & 0xff, DT_DATA_IN);

    return 0;
}

int dt2801_writedata2(unsigned int data)
{

    dprintk(( "dt2801: writedata2 data = %d.\n", data ));

    if ( dt2801_writedata( data & 0xff) != 0) {
        return -1;
    }
    if ( dt2801_writedata( (data >> 8) & 0xff ) != 0) {
        return -1;
    }

    return 0;
}

int dt2801_writecommand(int command)
{
    int stat;

    dprintk(( "dt2801: writecommand cmd = %d.\n", command ));

    if (dt2801->fake != 0) {
        return 0;
    }

    stat = inb_p(DT_STATUS);
    dprintk(( "dt2801: writecommand-a stat = %d.\n", stat ));

    stat = ((stat)^DT_S_DATA_IN_FULL) & DT_S_DATA_IN_FULL;
    dprintk(( "dt2801: writecommand-a stat = %d.\n", stat ));

    if (!stat) {
        dt2801_wait();
        dprintk(( "dt2801: writecommand-a back.\n" ));

        stat = inb_p(DT_STATUS);
        dprintk(( "dt2801: writecommand-a1 stat = %d.\n", stat ));
        stat = ((stat)^DT_S_DATA_IN_FULL) & DT_S_DATA_IN_FULL;
        dprintk(( "dt2801: writecommand-a1 stat = %d.\n", stat ));

        if (!stat) {
            dprintk("dt2801: writecommand timed out. DataInFull not cleared\n");
        return -1;
        }
    }

    dprintk(( "dt2801: writecommand-a ok.\n" ));

    stat = inb_p(DT_STATUS);
    dprintk(( "dt2801: writecommand-b stat = %d.\n", stat ));

    stat = stat & DT_S_READY;
    dprintk(( "dt2801: writecommand-b stat = %d.\n", stat ));

    if (!stat) {
        dt2801_wait();

        stat = inb_p(DT_STATUS);
        stat = stat & DT_S_READY;

        if (!stat) {
            dprintk("dt2801: writecommand timed out. Ready not set\n");
            return -1;
        }

    }

    dprintk(( "dt2801: writecommand ok cmd = %d.\n", command ));
    outb_p(command, DT_COMMAND);
    return 0;
}

int dt2801_do_reset(void)
{
    int stat, ignored;

    stat = dt2801_writecommand(DT_C_RESET);
    stat = dt2801_readdata(&ignored);

    return 0;
}

int dt2801_reset(void)
{
    int stat;


    dprintk(( "dt2801: reset.\n" ));

    stat = dt2801_do_reset();

    /* default after RESET */

    /* not yet existent
    dt2801_adc_reset();
    */
    dt2801_dac_reset();
    dt2801_dio_reset();

    /* clock (ticks !!) */   /* ??? */
    dt2801->clock = 400;

    dprintk(( "dt2801: reset done.\n" ));

    return 0;
}

static int dt2801_setclock(unsigned int period)
{

    dprintk(( "dt2801_setclock: %d.\n", period ));
    dt2801_writecommand(DT_C_SET_CLOCK);
    dt2801_writedata2(period);
    dt2801_checkforerror();   /* ??? */
    return dt2801_checkforerror();
}

int dt2801_ioctl(struct inode *inode, struct file *file,
                       unsigned int cmd, unsigned long arg)
{
    dprintk(( "dt2801_ioctl: %d.\n", cmd ));

    switch (cmd)
    {
        case DT2801_IOCTL_STOP: {
                dt2801_stop();
            }
            break;

        case DT2801_IOCTL_SETCLOCK: {
                int cl;
                cl = (int) arg;
                dprintk(( "dt2801: ioctl_setclock arg %d.\n", cl ));

                if ( dt2801_setclock(cl) != 0)
                    return -1;
                dt2801->clock = cl;

            }
            break;
        case DT2801_IOCTL_SETEXTCLOCK: {
                int extcl;
                extcl = (int) arg;
                dprintk(( "dt2801: ioctl_setextclock arg %d.\n", extcl ));

                dt2801->extclock = extcl;
            }
            break;
        case DT2801_IOCTL_SETEXTTRIGGER: {
                int exttrig;
                exttrig = (int) arg;
                dprintk(( "dt2801: ioctl_setexttrigger arg %d.\n", exttrig ));

                dt2801->exttrigger = exttrig;
            }
            break;
        case DT2801_IOCTL_SETDMA: {
                int dma_flag;
                dma_flag = (int) arg;
                dprintk(( "dt2801: ioctl_setdma arg %d.\n", dma_flag ));

                dt2801->dmamode = dma_flag;
            }
            break;
        case DT2801_IOCTL_SETCONTINOUS: {
                int cont_flag;
                cont_flag = (int) arg;
                dprintk(( "dt2801: ioctl_setcontinous arg %d.\n", cont_flag ));

                dt2801->continous = cont_flag;
            }
            break;

        default:
            printk("dt2801: unknown ioctl %u.\n", cmd);
            return -EINVAL;
    }

    return 0;
}

static int release_resources(void)
{
   /* Free resources */

    dt2801_adc_cleanup();
    dt2801_dac_cleanup();
    dt2801_dio_cleanup();

    free_dma(dt2801->dmachannel);
    release_region(dt2801->baseaddr, DT2801_IOSIZE);

    return 0;
}


/* called by the dt281_xxx_release */
void dt2801_release(struct inode *inode, struct file *file)
{
    dt2801->inuse = 0;   /* XXX */

    #ifdef MODULE
        MOD_DEC_USE_COUNT;
    #endif

    dprintk( "dt2801_release: DT2801 released.\n" );
}

int dt2801_open(struct inode *inode, struct file *file)
{
    int stat = 0;
    int minor = MINOR(inode->i_rdev);


    dprintk(( "dt2801_open: minor = %d.\n", minor ));


    if ( DT2801_ADC == minor )
    {
        dprintk(( "dt2801_open: ADC.\n" ));
        if (!dt2801->inuse) {
            file->f_op = &dt2801_adc_fops;
            stat = file->f_op->open(inode, file);
        } else {
            dprintk(( "dt2801_open: ADC open failed, device busy.\n" ));
            stat = -(EBUSY);
        }

    } else if ( DT2801_DAC0 <= minor && minor < DT2801_DAC0+DT2801_NUM_DACS) {

        if (!dt2801->inuse) {
            dprintk(( "dt2801_open: DAC.\n" ));
            file->f_op = &dt2801_dac_fops;
            stat = file->f_op->open(inode, file);
        } else {
            dprintk(( "dt2801_open: DAC open failed, device busy.\n" ));
            stat = -(EBUSY);
        }

    } else if ( DT2801_DIO0 <= minor && minor <= DT2801_DIO3) {

        dprintk(( "dt2801_open: DIO.\n" ));

	file->f_op = &dt2801_dio_fops;
    	stat = dt2801_dio_open(inode, file);

    } else {

        printk("dt2801_open: invalid minor.\n");
        return -(EINVAL);
    }

    if (stat < 0) {
        return stat;
    }


#ifdef MODULE
    MOD_INC_USE_COUNT;
#endif

    return stat;

}

struct file_operations dt2801_fops = {
    NULL,            /* seek */
    NULL,            /* read */
    NULL,            /* write */
    NULL,            /* readdir */
    NULL,            /* select */
    dt2801_ioctl,    /* ioctl */
    NULL,            /* mmap */
    dt2801_open,     /* open */
    dt2801_release   /* release */
};

#ifdef MODULE    /* only module version */

#if LINUX_VERSION_CODE <= 0x010300
char kernel_version[] = UTS_RELEASE;
#elif LINUX_VERSION_CODE >= 0x020000
#else
#error "check 'kernel_version' - symbol for your linux-version"
#endif

/*
 These can be set from the commandline, when loading the module.
*/
static int fake = 0;

static int major = DT2801_MAJOR;

static int io = DT2801_BASEADDR;
static int dma = DT2801_DMACHANNEL;

int init_module(void)
{

    printk("dt2801: DT2801 driver, %s\n", rcs_revision);

    if ( (dt2801 = (Dt2801_t *) kmalloc( sizeof( Dt2801_t), GFP_KERNEL ))
         == NULL) {
        return -(ENOMEM);
    }

    dt2801->fake = fake;

    dt2801->inuse = 0;


    dt2801->major      = major;
    dt2801->baseaddr   = io;
    dt2801->dmachannel = dma;


    dt2801->clock_min = 2;
    dt2801->clock_max = 0xffff;

    dt2801->inerror = 0;

    if ( request_dma(dt2801->dmachannel, "dt2801") ) {
        printk("dt2801_init: Cannot request DMA %d.\n", dt2801->dmachannel);
        goto fail1;
    }

    if (check_region(dt2801->baseaddr, DT2801_IOSIZE) != 0) {
        printk("dt2801: init_module: I/O ports 0x%lx to 0x%lx already in use.\n",
                   dt2801->baseaddr, dt2801->baseaddr + DT2801_IOSIZE-1);
        goto fail2;
    }

    request_region(dt2801->baseaddr, DT2801_IOSIZE, "dt2801");

    if (register_chrdev(dt2801->major, "dt2801", &dt2801_fops)) {
        printk("dt2801_init: Cannot register to major device %d.\n",
                         dt2801->major);
        goto error;
    } 


    dt2801_stop();  /* Stop to make sure the board can accept the next command */

    /* A succesfull reset() makes us believe the board is present */
    if (dt2801_do_reset() != 0) {
        printk("dt2801_init: board not detected !?\n");
        goto error;
    }


    /* Initialize A/D, D/A, D-I/O */
    dt2801->dmamode    = 0;
    dt2801->continous  = 0;
    dt2801->extclock   = 0;
    dt2801->exttrigger = 0;


    if (dt2801_adc_init() != 0) {
        printk("dt2801_adc_init: failed\n");
        goto error;
    }
    if (dt2801_dac_init() != 0) {
        printk("dt2801_dac_init: failed\n");
        goto error;
    }
    if (dt2801_dio_init() != 0) {
        printk("dt2801_dio_init: failed\n");
        goto error;
    }

    /* Again reset the board, this time via dt2801_reset(), which will call the
       dt2801_xxx_reset() routines.
    */
    if (dt2801_reset() != 0) {
        printk("dt2801_init: reset not ok.\n");
        goto error;
    }


    printk("dt2801: installed.\n");
    printk("dt2801: init_module: fake(%d), major(%d), dmachannel(%d), baseaddr(0x%lx).\n",
            dt2801->fake, dt2801->major, dt2801->dmachannel, dt2801->baseaddr);

    return 0;

error:

    printk("dt2801: cleaning up.\n");
    dt2801_adc_cleanup();
    dt2801_dac_cleanup();
    dt2801_dio_cleanup();

fail2:
    printk("dt2801: unloading driver.\n");
    release_region(dt2801->baseaddr, DT2801_IOSIZE);
fail1:
    free_dma(dt2801->dmachannel);

    kfree(dt2801);

    return -(EIO);
}

void cleanup_module(void)
{
    printk("dt2801: cleanup_module called.\n");

    /* As 'cleanup_mmodule()' should only be called when the module is no
       longer in use, this is not strictly necessary
    */
    if (MOD_IN_USE)
    {
        printk("dt2801: device busy, remove delayed\n");
        return;
    } else {

        release_resources();

        if (unregister_chrdev(dt2801->major, "dt2801") != 0)
        {
            printk("dt2801: cleanup_module failed\n");
        } else {
            printk("dt2801: cleanup_module succeeded\n");
        }
    }

    kfree(dt2801);
}

#endif /* MODULE */
